/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package stats;

import java.io.IOException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Logger;
import java.lang.Runtime;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;

/**
 *
 * @author Peter
 * Let's answer two questions with this small program.
 * What would Goldilocks use for the "just right" number of threads in a system application?
 * ANSWER: base it on the number of CPU's  (i.e. processors) running on the machine.
 * How can we do some accounting and create some basic statistics about the threads we run? 
 * ANSWER: extend the ThreadPoolExecutor class. Fortunately, or, unfortunately, this means taking
 *         another look at the ThreadPoolExecutor() class.
 * How can we write data using the Logger facility? 
 * ANSWER: two ways - to the "output screen" and to a log. The log should be associated with
 *         a file handler.
 * Umm, Let's answer three questions with this small program.
 * How do we use atomic variables?
 * ANSWER: with Atomicxxx data types. AtomicLong in the case below.
 * ok, we have a bunch of questions that can be answered with this program.
 * One more: What is threadlocal??
 * ANSWER: It is memory/storage available for threads to share, where the jvm will do the
 *         syncrhonization for us.
 */
public class ThreadPoolAccountant extends ThreadPoolExecutor {
    private final ThreadLocal <Long> startTime = new ThreadLocal <Long>();
    private final AtomicLong numTasks = new AtomicLong();
    private final AtomicLong totalTime = new AtomicLong();        
    private final Logger logg;            

    private Handler handler;     
    public ThreadPoolAccountant(int corePoolSize, int maximumPoolSize, long keepAliveTime, TimeUnit unit, BlockingQueue<Runnable> workQueue) {
        //from the javadoc:
        // corePoolSize - the number of threads to keep in the pool, even if they are idle
        // maximumPoolSize - the maximum number of threads to allow in the pool
        // keepAliveTime - when the number of threads is greater than the core, 
        //                 this is the maximum time that excess idle threads will wait for new tasks before terminating.
        // unit - the time unit for the keepAliveTime argument
        // workQueue - the queue to use for holding tasks before they are executed. 
        //             This queue will hold only the Runnable tasks submitted by the execute method.
        super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue); 
        logg = Logger.getLogger("");
        try {
            // the maximum number of bytes to write to any one file
            int limit = 5 * 1024 * 1024; 
            // the number of files to use
            int count = 10;              
            handler = new FileHandler("log_test.txt", limit, count);            
            
            logg.addHandler(handler);
            //Logger.getLogger("com.wombat").setLevel(Level.FINEST);
            int numberOfCPUs;
            numberOfCPUs = Runtime.getRuntime().availableProcessors();
            //from: JAVA Concurrency in Practice, by Brian Goetz)
            //let's avoid too many or too few threads.
            //number of threads N = number of cpus * u * (1 + w/c)
            //where u = target cpu utilization 0 < u < 1
            //where w/c = ratio of wait time to compute time, let's set it to 1.
            //for purposes of this exercise: N = number of cpus * 0.8 *(1+1)
            //and... corePoolSize = N;
            logg.info(String.format("Number of Availabe Processors: %d",numberOfCPUs));
            //logg.fine(String.format("Number of Availabe Processors: %d",numberOfCPUs));
        } catch (IOException ex) {
            Logger.getLogger(ThreadPoolAccountant.class.getName()).log(Level.SEVERE, null, ex);
        } catch (SecurityException ex) {
            Logger.getLogger(ThreadPoolAccountant.class.getName()).log(Level.SEVERE, null, ex);
        }        
    }
    
    protected void beforeExecute(Thread t, Runnable r) {
        super.beforeExecute(t, r);
        logg.info(String.format("Thread %s: start %s", t, r));
        //logg.fine(String.format("Thread %s: start %s", t, r));
        System.out.println("beforeExecute: " + t + " thread\n");
        startTime.set(System.nanoTime());
    }
    
    protected void afterExecute(Runnable r, Throwable thr) {
        try {
            long endTime = System.nanoTime();
            long taskTime = endTime - startTime.get();
            numTasks.incrementAndGet();
            totalTime.addAndGet(taskTime);
            double seconds = (double)taskTime  / 1000000000.0;
            logg.info(String.format("Thread %s: end %s, elapsed time = %d seconds", thr, r, seconds));
        } finally {super.afterExecute(r, thr); }
    }
    
    protected void terminated() {
        try {
            logg.info(String.format("Terminate: average time = %dns", totalTime.get() / numTasks.get()));
            double seconds = (double)totalTime.get() / 1000000000.0;
            logg.info(String.format("Terminate: average time = " + seconds / numTasks.get() + " seconds."));
            //logg.fine(String.format("Terminate: average time = %dns", totalTime.get() / numTasks.get()));
            //double seconds = (double)totalTime.get() / 1000000000.0;
            //logg.fine(String.format("Terminate: average time = %d seconds", seconds / numTasks.get()));
        } finally {super.terminated(); }        
    }    
} // end class ThreadPoolAccountant
